//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"context"
	"net/url"
	"sync/atomic"
	"time"
)

type Job struct {
	im        *Importer
	ctx       context.Context
	cancel    context.CancelFunc
	State     JobState  `json:"state"`
	CreatedAt time.Time `json:"created"`
	ID        uint64    `json:"id"`
	Group     string    `json:"group"`
	User      string    `json:"user"`
	Source    SourceURL `json:"source"`
	RefID     string    `json:"ref-id,omitempty"`
}

type Jobs []Job

func (job *Job) run() (err error) {
	atomic.StoreUint32((*uint32)(&job.State), uint32(JobRunning))
	defer job.cleanup()

	// TODO: update store.file: set import state to ImportRunning

	// do the actual import here
	// use job.ctx.Done() wherever we wait for something
	// return ErrTimeout in case job.ctx.Done() is closed
	job.im.dbgLog.Printf("running import for %s/%d from: %s", job.Group, job.ID, job.Source.String())
	time.Sleep(60 * time.Second)
	// update store.file: set import state to ImportDone or ImportAborted
	// send result to all done subscriptions
	// trigger removal from job inventory

	return
}

func (job *Job) start() <-chan error {
	done := make(chan error, 1)
	go func() {
		done <- job.run()
		close(done)
	}()
	return done
}

// handle subscriptions for progress and done

func (job *Job) Start(timeout time.Duration) error {
	if !atomic.CompareAndSwapUint32((*uint32)(&job.State), uint32(JobNew), uint32(JobInitializing)) {
		return nil
	}

	// shall we use a global context here so we can cancel all jobs at once? i.e. for forced shutdown...
	if timeout > 0 {
		job.ctx, job.cancel = context.WithTimeout(context.Background(), timeout)
	} else {
		job.ctx, job.cancel = context.WithCancel(context.Background())
	}

	select {
	case job.im.work <- job:
	default:
		job.cleanup()
		return ErrTooManyJobs
	}
	// TODO: update store.file: set import state to ImportPending
	atomic.StoreUint32((*uint32)(&job.State), uint32(JobPending))
	return nil
}

func (job *Job) Cancel() {
	if atomic.LoadUint32((*uint32)(&job.State)) < uint32(JobPending) {
		return
	}
	job.cancel()
}

func (job *Job) cleanup() {
	atomic.StoreUint32((*uint32)(&job.State), uint32(JobDestroying))
	// we are not really interested in the result here
	go job.im.deleteJob(job.Group, job.ID)
}

func newJob(im *Importer, group string, id uint64, src url.URL, user, refID string) *Job {
	job := &Job{im: im, Group: group, ID: id, Source: SourceURL(src), User: user, RefID: refID}
	job.CreatedAt = time.Now()
	return job
}
