//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"log"
)

type listJobsResp struct {
	jobs Jobs
	err  error
}

type listJobsReq struct {
	resp  chan listJobsResp
	group string
}

type insertOrGetJobResp struct {
	job *Job
	err error
}

type insertOrGetJobReq struct {
	resp  chan insertOrGetJobResp
	group string
	id    uint64
	job   *Job
}

type getJobResp struct {
	job *Job
	err error
}

type getJobReq struct {
	resp  chan getJobResp
	group string
	id    uint64
}

type deleteJobResp struct {
	err error
}

type deleteJobReq struct {
	resp  chan deleteJobResp
	group string
	id    uint64
}

type jobInventoryGroup struct {
	jobs map[uint64]*Job
	// add list of subscription channels
}

func newJobInventoryGroup() *jobInventoryGroup {
	g := &jobInventoryGroup{}
	g.jobs = make(map[uint64]*Job)
	return g
}

type jobInventory struct {
	infoLog         *log.Logger
	errLog          *log.Logger
	dbgLog          *log.Logger
	listJobsC       chan listJobsReq
	insertOrGetJobC chan insertOrGetJobReq
	getJobC         chan getJobReq
	deleteJobC      chan deleteJobReq
	groups          map[string]*jobInventoryGroup
}

func (i *jobInventory) listJobs(group string) (resp listJobsResp) {
	ig, exists := i.groups[group]
	if !exists {
		return
	}

	for _, j := range ig.jobs {
		resp.jobs = append(resp.jobs, *j)
	}
	return
}

func (i *jobInventory) insertOrGetJob(group string, id uint64, job *Job) (resp insertOrGetJobResp) {
	ig, exists := i.groups[group]
	if !exists {
		ig = newJobInventoryGroup()
		ig.jobs[id] = job
		i.groups[group] = ig
		resp.job = job
		return
	}

	if resp.job, exists = ig.jobs[id]; !exists {
		ig.jobs[id] = job
		resp.job = job
		return
	}

	i.dbgLog.Printf("importer: job-inventory added job(%s/%d)", group, id)
	return
}

func (i *jobInventory) getJob(group string, id uint64) (resp getJobResp) {
	ig, exists := i.groups[group]
	if !exists {
		resp.err = ErrNotFound
		return
	}

	if resp.job, exists = ig.jobs[id]; !exists {
		resp.err = ErrNotFound
		return
	}

	return
}

func (i *jobInventory) deleteJob(group string, id uint64) (resp deleteJobResp) {
	ig, exists := i.groups[group]
	if !exists {
		resp.err = ErrNotFound
		return
	}
	if _, exists = ig.jobs[id]; !exists {
		resp.err = ErrNotFound
		return
	}

	delete(ig.jobs, id)
	i.dbgLog.Printf("importer: job-inventory removed job(%s/%d)", group, id)
	if len(ig.jobs) == 0 {
		// TODO: only do this if there are no subscriptions
		delete(i.groups, group)
		i.dbgLog.Printf("importer: job-inventory also removed now empty group %s", group)
	}
	return
}

func (i *jobInventory) dispatchRequests() {
	i.dbgLog.Println("importer: job-inventory initialized")
	for {
		select {
		case req := <-i.listJobsC:
			req.resp <- i.listJobs(req.group)
		case req := <-i.insertOrGetJobC:
			req.resp <- i.insertOrGetJob(req.group, req.id, req.job)
		case req := <-i.getJobC:
			req.resp <- i.getJob(req.group, req.id)
		case req := <-i.deleteJobC:
			req.resp <- i.deleteJob(req.group, req.id)
		}
	}
}

// *********************************************************
// Public Interface

func (i *jobInventory) ListJobs(group string) (Jobs, error) {
	respC := make(chan listJobsResp)
	req := listJobsReq{respC, group}
	i.listJobsC <- req
	resp := <-respC
	return resp.jobs, resp.err
}

func (i *jobInventory) InsertOrGetJob(group string, id uint64, job *Job) (*Job, error) {
	respC := make(chan insertOrGetJobResp)
	req := insertOrGetJobReq{respC, group, id, job}
	i.insertOrGetJobC <- req
	resp := <-respC
	return resp.job, resp.err
}

func (i *jobInventory) GetJob(group string, id uint64) (*Job, error) {
	respC := make(chan getJobResp)
	req := getJobReq{respC, group, id}
	i.getJobC <- req
	resp := <-respC
	return resp.job, resp.err
}

func (i *jobInventory) DeleteJob(group string, id uint64) error {
	respC := make(chan deleteJobResp)
	req := deleteJobReq{respC, group, id}
	i.deleteJobC <- req
	resp := <-respC
	return resp.err
}

// TODO: handle subscriptions to new and deleted jobs

func newJobInventory(infoLog, errLog, dbgLog *log.Logger) *jobInventory {
	i := &jobInventory{infoLog: infoLog, errLog: errLog, dbgLog: dbgLog}
	i.listJobsC = make(chan listJobsReq)
	i.insertOrGetJobC = make(chan insertOrGetJobReq)
	i.getJobC = make(chan getJobReq)
	i.deleteJobC = make(chan deleteJobReq)
	i.groups = make(map[string]*jobInventoryGroup)

	go i.dispatchRequests()
	return i
}
